package cn.zhangdeman.exception;

import cn.zhangdeman.exception.annotation.CustomExceptionScanAnnotation;
import jakarta.servlet.ServletContextEvent;
import jakarta.servlet.ServletContextListener;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.stereotype.Component;
import org.springframework.util.ClassUtils;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;


// 基于hashmap实现的内存缓存
// 因为就是服务启动时初始化一次, 无需考虑多线程并发安全问题
@Component
public class HashMapCache implements ServletContextListener {
    // 单例: 饿汉模式
    private static final HashMapCache hashMapCache = new HashMapCache();
    private final Map<String, ICustomExceptionConfig> cache = new HashMap<>();
    @Autowired
    private ApplicationContext applicationContext;
    //将构造器设置为private禁止通过new进行实例化
    private HashMapCache() {

    }
    // 此处实现的是自定义注解的解析逻辑, 将错误码全部加载到内存中
    public void loadCustomException() throws ClassNotFoundException, InvocationTargetException, IllegalAccessException, NoSuchMethodException, InstantiationException {
        Map<String, Object> customExceptionScanAnnotationMap = applicationContext.getBeansWithAnnotation(CustomExceptionScanAnnotation.class);
        for(Object bean : customExceptionScanAnnotationMap.values()){
            Class<?> extensionClz = ClassUtils.getUserClass(bean);
            CustomExceptionScanAnnotation customExceptionAnnotation = AnnotationUtils.findAnnotation(extensionClz, CustomExceptionScanAnnotation.class);
            assert customExceptionAnnotation != null;
            String[] scanPackageList = customExceptionAnnotation.exceptionPackageList();
            String[] exceptionClassList = customExceptionAnnotation.exceptionClassList();
            List<String> classList = getAllClassList(scanPackageList, exceptionClassList);
            for (String className: classList) {
                Class<?> classInstance = Class.forName(className);
                // 判断是否实现接口: 要求实现 ICustomExceptionAnnotation 和 ICustomExceptionConfig
                if(!ICustomExceptionAnnotation.class.isAssignableFrom(classInstance) || !ICustomExceptionConfig.class.isAssignableFrom(classInstance)) {
                    // 未实现接口
                    continue;
                }
                // 实现了接口
                // 反射获取类实例
                if (!classInstance.isEnum()) {
                    // 不是枚举类
                    Constructor<?> constructor = classInstance.getDeclaredConstructor();
                    ICustomExceptionAnnotation iCustomExceptionAnnotation = (ICustomExceptionAnnotation) (constructor.newInstance());
                    List<ICustomExceptionConfig> list = iCustomExceptionAnnotation.list();
                    for (ICustomExceptionConfig iCustomExceptionConfig: list) {
                        getInstance().cache.put(iCustomExceptionConfig.getCategory()+"_"+iCustomExceptionConfig.getCode(), iCustomExceptionConfig);
                    }
                } else {
                    // 是枚举类
                    Method method = classInstance.getMethod("values");
                    Object[] objectList = (Object[]) method.invoke(null);
                    for (Object itemEnumObject: objectList) {
                        ICustomExceptionConfig iCustomExceptionConfig = (ICustomExceptionConfig)itemEnumObject;
                        getInstance().cache.put(iCustomExceptionConfig.getCategory()+"_"+iCustomExceptionConfig.getCode(), iCustomExceptionConfig);
                    }
                }
            }
        }
    }
    // 获取实例
    public static HashMapCache getInstance() {
        return hashMapCache;
    }

    // 读取描述
    public String getMessage(String category, String code) {
        String key = category + "_" + code;
        if (!getInstance().cache.containsKey(key)) {
            // 不包含指定code
            return key;
        }
        return getInstance().cache.get(key).getMessage();
    }

    // 添加
    public void put(String category, String code, String message) {
        String key = category + "_" + code;
        getInstance().cache.put(key, new CustomExceptionConfig(category, code, message));
    }


    // 直接添加异常实例
    public void putByConfig(ICustomExceptionConfig iCustomExceptionConfig) {
        String key = iCustomExceptionConfig.getCategory() + "_" + iCustomExceptionConfig.getCode();
        getInstance().cache.put(key, iCustomExceptionConfig);
    }

    // 获取一个包下的所有类列表
    private List<String> getPackageClassList(String packageName) {
        List<String> classList = new ArrayList<>();
        String path = packageName.replace(".", "/");
        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
        File directory = new File(Objects.requireNonNull(classLoader.getResource(path)).getFile());
        if (!directory.exists()) {
            return classList;
        }
        File[] fileList = directory.listFiles();
        assert fileList != null;
        for (File file : fileList) {
            String fileName = file.getName();
            if (!fileName.endsWith(".class")) {
                // 不是类文件
                continue;
            }
            // 移除.class后缀
            String className = packageName + "." + fileName.substring(0, fileName.length() - 6);
            classList.add(className); // 添加到列表
        }
        return classList;
    }

    // 同时获取多个包下的class列表, 并支持重复的数据去重
    private List<String> getClassListByPackageList(String[] packageList) {
        List<String> classList = new ArrayList<>();
        for (String packageName: packageList) {
            List<String> packageClassList = getPackageClassList(packageName);
            for (String className: packageClassList) {
                if (classList.contains(className)) {
                    // 已存在, 无需重复添加
                    continue;
                }
                classList.add(className);
            }
        }
        return classList;
    }

    // 获取输入的全部class列表
    private List<String> getAllClassList(String[] packageList, String... configClassNameList) {
        List<String> classList =getClassListByPackageList(packageList);
        for (String className: configClassNameList) {
            if (classList.contains(className)) {
                // 已存在, 无需重复添加
                continue;
            }
            classList.add(className);
        }
        return classList;
    }

    @Override
    public void contextInitialized(ServletContextEvent sce) {
        System.out.println("[start]自动执行异常配置注入" + new Date());
        try {
            this.loadCustomException();
        } catch (ClassNotFoundException e) {
            throw new RuntimeException(e);
        } catch (InvocationTargetException e) {
            throw new RuntimeException(e);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        } catch (NoSuchMethodException e) {
            throw new RuntimeException(e);
        } catch (InstantiationException e) {
            throw new RuntimeException(e);
        }
        System.out.println("[finish]自动执行异常配置注入" + new Date());
    }

    @Override
    public void contextDestroyed(ServletContextEvent sce) {
    }
}

